/* $Id$ */

/*
 *
 *  CS213 - Lab assignment 3
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>

#include "memlib.h"
#include "malloc.h"


team_t team = {
    /* Team name to be displayed on webpage */
    "Who's John Galt?",
    /* First member full name */
    "Blake Scholl",
    /* First member email address */
    "bscholl",
    /* Second member full name (leave blank if none) */
    "Stephanie Song",
    /* Second member email address (blank if none) */
    "wsong"
};


/* Fixme: Figure out what this should be */
#define MAX_WASTE 16 

/* function prototype */

void printchunk(chunkptr);



/*
   An allocated chunk looks like this:


    chunk-> +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Size of previous chunk, if allocated            | |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Size of chunk, in bytes                         |P|
      mem-> +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             User data starts here...                          .
            .                                                               .
            .             (malloc_usable_space() bytes)                     .
            .                                                               |
nextchunk-> +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Size of chunk                                     |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	    
	    **The size of a chunk DOES NOT include header and footers**
	    ** P is set to 1 if the chunk is allocated, 0 if the chunk is free **


   Free chunks are stored in circular doubly-linked lists, and look like this:

    chunk-> +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Size of previous chunk                            |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    `head:' |             Size of chunk, in bytes                         |P|
      mem-> +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Forward pointer to next chunk in list             |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Back pointer to previous chunk in list            |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |             Unused space (may be 0 bytes long)                .
            .                                                               .
            .                                                               |
nextchunk-> +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    `foot:' |             Size of chunk, in bytes                           |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

*/

/************************************************************************
 *                   G L O B A L   V A R I A B L E S  
 ************************************************************************/

/* declare the bin(s).  
 *
 * Currently, we are lame and have only 1.
 *
 */

chunkptr binrest; /* store chunks of any size */


/* keep track of the largest free chunk around,
 * so when the heap needs to be grown, we 
 * don't waste time poking around for a chunk when there isn't one 
 */

/* FIXME: Implement this feature 

  we will need additional code to keep track of the largest stuff in each bin,
  and bin insert/remove routines will need to maintain these data

*/

int largestFreeChunk; 

/* The system page size, in bytes  */
int page_size;


/************************************************************************
 *                          F U N C T I O N S
 ************************************************************************/

/*
 * Return a int pointer to the footer of a chunk
 */
int * foot (chunkptr c)
{
 
  char * retval = (char  *) c;
  retval = retval + 8 + (int) (0xFFFFFFFE & c->size);
  return (int *) retval;
}


/*
 * Print information about a chunk, doing tests to make sure everything is kosher
 */
void printchunk(chunkptr p)
{

  printf("Chunk at 0x%x %s:\n", (int) p, (p->size & 0x1) ? "(allocated)" : "(free)" );
  printf("\tPrevious physical chunk size: 0x%x \n",  p->prev_size);
  printf("\tSize: 0x%x (%d) \n", p->size, p->size);
  printf("\tFooter (at 0x%x) reports size: 0x%x \n", (int) foot(p), *foot(p));
  if((char *) p > dseg_hi || (char *) p < dseg_lo)
    printf("\t**chunk is outside of heap.\n");
  if(((char * ) foot(p) > dseg_hi) || ((char * ) foot(p) < dseg_lo))
    printf("\t**chunk's footer is outside of heap.\n");

  if(0x7 & (int)p)
    printf("\t**chunk is not aligned properly.\n");
  
  if(0x7 & (p->size & 0xFFFFFFFE))
    printf("\t** chunk's size is not a multiple of 8 bytes!\n");

  if(*foot(p) != p->size)
    printf("\t** chunk's size and footer do not match!\n");
      
  if(p->size == 0)
    printf("\t** chunk has 0 size!\n");

}



/*
 * Run some checks on a chunk, verifying that everything is kosher with it.  
 * Return 1 if there is a problem.
 */

int checkchunk(chunkptr p)
{

  if((char *) p > dseg_hi || (char *) p < dseg_lo)
    return 1;
  if(((char * ) foot(p) > dseg_hi) || ((char * ) foot(p) < dseg_lo))
    return 1;

  if(0x7 & (int)p)
    return 1;
  
  if(0x7 & (p->size & 0xFFFFFFFE))
    return 1;
 
  if(*foot(p) != p->size)
    return 1;
  
  if(p->size==0)
    return 1;

  return 0;

}




/*
 * Walk along the bins, printing out info about each chunk 
 */
void binwalk()
{
  chunkptr trav = binrest;
  int first = 1;


  int lastsize;

  printf("************ BINWALK ************\n");
  if(!trav)
    {
      printf("Bin empty.\n");
      return;
    }
  
  lastsize = trav->size;
  while(trav && ((trav != binrest) | first))
    {
      if(trav->size > lastsize)
	{
	  printf("** ERROR: bin is not sorted properly!\n");
	}
      lastsize = trav->size;
      printchunk(trav);

      trav= trav->next;
      first=0;
    }
  printf("*********************************\n");
}


/* 
 * Run checks on all blocks inside the address space
 */
int memwalk_check()
{

  chunkptr trav = (chunkptr) dseg_lo;

  while(trav < (chunkptr) ((char *) dseg_hi -3 ))
    {
      if( checkchunk(trav) )
	{
	  printchunk(trav);
	  return 1;
	}

      trav = (chunkptr) foot(trav);
    }


  return 0;

}




/* Print out info about all the bins */
int binwalk_check()
{
  chunkptr trav = binrest;
  int first = 1;


  int lastsize;

  /* Check all the blocks in memory.  */
  if(memwalk_check())
    {
      printf("** ERROR:  Zikes, there was a problem with memory!\n");
      binwalk();
      while(1);

      return 1;
    }



  if(!trav)
    {
      return 0;
    }
  
  lastsize = trav->size;
  while(trav && ((trav != binrest) | first))
    {
      if(trav->size > lastsize)
	{
	  
	  printf("** ERROR: (binwalk_check) bin is not sorted properly!!\n");
	  binwalk();
	  return 1;
	}

      if(checkchunk(trav))
	{
	  printf("** Error: (binwalk_check) problem with a chunk!\n");
	  binwalk();
	  return 1;
	}

      lastsize = trav->size;

      trav= trav->next;
      first=0;
    }

  return 0;
}


/* 
 * Insert the given chunk into the list
 *
 * The circularly list is sorted in decreasing order; for ties, give
 * preference to blocks already in the list
 * 
 * The chunkptr list must be the first node in the list; e.g. the one with the largest
 * size.
 *
 */

void  bin_insert_chunk(chunkptr *list, chunkptr chunk)
{

  chunkptr trav = (*list)->next;

#ifdef DEBUG

  if(!list)
    {
      fprintf(stderr, "%s:%d bin_insert_chunk error: NULL list", __FILE__, __LINE__);
      return;
    }

  if(!chunk)
    {
      fprintf(stderr, "%s:%d bin_insert_chunk error: NULL chunk", __FILE__, __LINE__);
      return;
    }

#endif

  /* scoot trav along the list until it points to the first chunk that is smaller than 
     the chunk we are inserting.  If we go all the way around the list, insert at the end of the list 
     (e.g. before the start of the list) */

  /*FIXME: Is this correct when we insert into the head of the list? ??? */

  /* check to see if there is only one node in the list */
  if(trav->next == trav)
    {
      /* insert after the front node */
      if(trav->size > chunk->size)
	{
	  trav->next = trav->prev = chunk;
	  chunk->next = chunk->prev = trav;

#ifdef DEBUG_CHECKBIN
	  if(binwalk_check()) 
	    {
	      printf("** ERROR: Bin boogered after inserting into bin (case 0)\n");
	      exit(0);
	    }
#endif

	  return;
	}
      else
	{
	  /* The chunk we are inserting is larger than the head chunk,
	   *  so replace the head/front node with the new chunk
	   */
	  trav->next = trav->prev = chunk;
	  chunk->next = chunk->prev = trav;
	  *list = chunk;

#ifdef DEBUG_CHECKBIN
	  if(binwalk_check()) 
	    {
	      printf("** ERROR: Bin boogered after inserting into bin (case 1)\n");
	      exit(0);
	    }
#endif

	  return;
	}
    }
  
  /* insert into a list with more than 1 element */


  /* First, check to see if the new chunk belongs at the head */
  if( chunk->size > (*list)->size)
    {

      /* Put it at the head of the list and be done with it, dammit */

      trav = (*list);

      chunk->next = trav;
      chunk->prev = trav->prev;
      chunk->prev->next = chunk;
      trav->prev = chunk;
      *list = chunk;

#ifdef DEBUG_CHECKBIN
      if(binwalk_check()) 
	{
	  printf("** ERROR: Above trace after inserting into bin (case 2)\n");
	  printf("\tOccured while inserting ");
	  printchunk(chunk);
	  exit(0);
	}
#endif


      return;
    }

  
  while((chunk->size < trav->size) && trav != *list)
    trav = trav->next;

  /* Now, insert _BEFORE_ trav */
  chunk->prev = trav->prev;
  chunk->prev->next = chunk;
  chunk->next = trav;
  trav->prev = chunk;

#ifdef DEBUG_CHECKBIN
  if(binwalk_check()) 
    {
      printf("** ERROR: Above trace after inserting into bin (case 3)\n");
      exit(0);
    }

#endif


}


/* Remove the best fit chunk from the bin.
 * If it is way too big, it is the job of the client to split the block
 * and add it to the proper bin 
 *
 * A pointer to the bin is passed in, so that when 
 * the head chunk is removed, the pointers are updated properly
 */

chunkptr bin_get_best_fit(chunkptr* bin, int size)
{
  
  chunkptr front = *bin;

  int * footer;

  /* is there only one thing in the bin? */
  if(front->next == front)

    {
      /* Is the only chunk available big enough? */
      if(front->size >= size)
	{

#ifdef DEBUG_VERY
	  printf("The only thing in the bin is big enough, allocating it.\n");
#endif

	  /* mark the bin as empty */
	  *bin = NULL;
	  
	  /* mark the chunk as used */
	  front->size = front->size | 0x1;
	  footer = foot(front);
	  *footer = front->size;

#ifdef DEBUG_VERY
	  printchunk(front);
#endif

#ifdef DEBUG_BINCHECK
	  if(binwalk_check()) 
	    {
	      printf("** ERROR: Bin boogered after getting best fit. (case 0) \n");
	      exit(0);
	    }
#endif
 
	  return front;
	}
      /* only chunk available is NOT big enough */

#ifdef DEBUG_BINCHECK
      if(binwalk_check()) 
	{
	  printf("** ERROR: Bin boogered after getting best fit. (case 1) \n");
	  exit(0);
	}
#endif

      return NULL;
    }
  else
    {
      /* More than one chunk in the bin, hunt for best fit */

      chunkptr trav = front->next;

      /* Check that there is a chunk in the bin which is large enough */
      
      if(front->size < size)
	{
	  
#ifdef DEBUG_BINCHECK
	  if(binwalk_check()) 
	    {
	      printf("** ERROR: Bin boogered after getting best fit. (case 2) \n");
	      exit(0);
	    }
#endif

	  
	  
	return 0;
	}

      
      while(( trav->size > size) && trav != front)
	{
	  trav = trav->next;
	}
      
      /* trav->prev is now the best fit.  Remove it from the list and return it */

      /* two cases now, trav could be the first chunk with EXACTLY
       * the size we want, or it could be the first that is too small.
       * 
       * so, if trav is too small, go back one in the list
       */

      if(trav->size < size) 
	trav = trav->prev;


      /* remove trav from the list */
      trav->prev->next = trav->next;
      trav->next->prev = trav->prev;

      /* mark trav as allocated */

      trav->size = trav-> size | 0x1;
      footer = foot(trav);
      *footer = trav->size;

      /* If we're removing the front of the list, we need to update the bin pointer */
      if(trav == *bin)
	{
#ifdef DEBUG_VERY
	  printf("bin_get_best_fit: Removed front of bin, updating bin pointer.\n");
#endif

	  *bin = trav->next;
	}
      
#ifdef DEBUG_BINCHECK
      if(binwalk_check()) 
	{
	  printf("** ERROR: Bin boogered after getting best fit (case 3)\n");
	  printchunk(trav);
	  exit(0);
	}
#endif

      return trav;
    }

}


//#define DEBUG

/* coalescing done here */
/* takes a pointer to a free block after flag fields are cleared */
/* and return a pointer to newly coalesced block */
chunkptr coalescing (chunkptr chunk)
{

  chunkptr tchunk;
  int size;
  

#ifdef DEBUG
  printf("<3<3<3<3<3<3<3<3 Coalescing <3<3<3<3<3<3<3<3<3<3<3\n");
  printchunk(chunk);
#endif
  
  /* check chunk in front */
  if (!((chunk->prev_size) & 0x1) && (chunk->prev_size != 0) )
    {
      
      /* chunk in front is free!!! Yeah!!! let's join :) */
      tchunk = (chunkptr) ((char*) chunk - chunk->prev_size - 8);
      
#ifdef DEBUG
      printf("Chunk free in front, join\n");
      printchunk(tchunk);
      if(checkchunk(tchunk))
	{
	  printf("** Coalescing: Boogered (%s:%d)\n", __FILE__, __LINE__);
	  exit(0);
	}
#endif
      
      /* relink pointers, if tchunk is the only thing in the bin, set bin
         to 0, otherwise, relink. */
      if (tchunk->next == tchunk)
	binrest = 0;
      else 
	{
	  if (tchunk == binrest)
	    binrest = tchunk->next;
	  
	  tchunk->next->prev = tchunk->prev;
	  tchunk->prev->next = tchunk->next;
	}
      
      size = tchunk->size + chunk->size + 8;
      tchunk->size = size;
      chunk = tchunk;
      
      /* set size at foot */
      *foot(chunk) = size;

#ifdef DEBUG
      if(checkchunk(tchunk))
	{
	  printf("** Coalescing: Boogered (%s:%d)\n", __FILE__, __LINE__);
	  exit(0);
	}
#endif

    }
  
  
  tchunk = (chunkptr) foot(chunk);
  
  
  /* we now take care of chunk in back */
  if ( !((tchunk->size) & 0x1) && (tchunk->size != 0) )
    {
      if ( (char*) foot(tchunk) <= dseg_hi - 3)
	{
#ifdef DEBUG
	  printf("chunk in back free\n");
	  printchunk(tchunk);
	  if(checkchunk(tchunk))
	    {
	      printf("** Coalescing: Boogered (%s:%d)\n", __FILE__, __LINE__);
	  exit(0);
	    }
#endif
	  
	  /* reset pointers for the chunk in back */
	  if (tchunk->next == tchunk)
	    binrest = 0;
	  else 
	    {
	      if (tchunk == binrest)
		binrest = tchunk->next;
	      
	      tchunk->next->prev = tchunk->prev;
	      tchunk->prev->next = tchunk->next;
	    }
      
	  
	  /* chunk in back is free, let's join */
	  size = chunk->size + tchunk->size + 8;
	  chunk->size = size;
	  *foot(chunk) = size;
	}
    }

#ifdef DEBUG
  printf("Return chunk!!!!!!!!\n");
  printchunk(chunk);
  printf("<3<3<3<3<3<3<3<3<3<3<3<3<3<3<3\n");
#endif
  
#ifdef BINWALK_CHECK

  if(checkchunk(chunk))
    {
      printf("** Coalescing returning boogered chunk!!\n");
      exit(0);
    }

  if(binwalk_check())
    {
      printf("** Coalescing boogered the memory!\n");
      exit(0);
    }
  
#endif

  return chunk;
  
}

  
/* put the given chunk in an appropriate bin
 * FIXME: Do coalescing here, or figure out lazy coalescing
 */
void chunk_free(chunkptr chunk)
{
  
  /* clear the allocated flag */
  chunk->size = chunk->size & 0xFFFFFFFE;
  *foot(chunk) = chunk->size;

  if(!binrest) 
    {
      /* the bin is empty, make the chunk the entirety of the bin */
      
      binrest = chunk;
      chunk->next = chunk;
      chunk->prev = chunk;

    }
  else
    {
      /* insert the chunk into the bin */
      bin_insert_chunk(&binrest, chunk);
    }

}


/* If the chunk wastes more than MAX_WASTE bytes, split it:
 * Size is the needed size of the block.
 * Invariant: size Is pre-adjusted to account for alignment issues
 * Return the allocated chunk, returning the newly split chunk to a bin (by calling chunk_free)
 *
 */

chunkptr conditional_split_chunk(chunkptr chunk, int size )
{
  int wasted_bytes = (chunk->size & 0xFFFFFFFE) - size;
 
  /* If we don't need to do any splitting, just return the chunk */
  if(wasted_bytes <= MAX_WASTE)

    {
#ifdef DEBUG_SPLIT
      printf("Wasting %d bytes, not splitting.\n", wasted_bytes);
#endif

      return chunk;
    }
  /* Split */
  {

    /* alignment of data. chunk is aligned already so we only have to
       make sure that size is a multiple of 8 */
    int newsize = (size + 7) & 0xFFFFFFF8;
    int remain_size = chunk->size - newsize - 8;
    
    chunkptr newchunk;
    
#ifdef DEBUG_SPLIT
    printf("Splitting, rather than wasting %d bytes.\n", wasted_bytes);
#endif

    chunk->size = newsize;
    



    /* create a new chunk -- the new chunk starts at the footer of the previous */
    newchunk = (chunkptr) foot(chunk);

    /* Set size and footer for the chunk we are returing to free */
    newchunk->size = (remain_size & 0xFFFFFFFE);
    *(foot(newchunk)) = newchunk->size;

#ifdef DEBUG_SPLIT
    printf("conditional-split:  new chunk size 0x%x, leftover size 0x%x\n",
	   newsize, remain_size);
#endif

    /* mark the chunk we are returning as allocated */
    chunk->size = chunk->size | 0x1;
    *foot(chunk) = chunk->size;



      /* Add the newly-created free chunk to a bin */

    chunk_free(newchunk);


#ifdef DEBUG_SPLIT
    printf("conditional-split allocated: ");
    printchunk(chunk);
    printf("conditional-split returned to bin: ");
    printchunk(newchunk);
#endif

    /* Return the reduced-size chunk */

#ifdef DEBUG_BINCHECK
     if(binwalk_check()) 
       {
	 printf("** ERROR: Bin boogered after conditional split!\n");
	 exit(0);
       }
#endif
     
    return chunk;

  }
}



/*
 *  Allocate the requested block of memory, or return NULL if there is 
 *  insufficient heap to do so
 */
void * chunk_alloc(int size)
{
  /* FIXME: Implement bin choosing in some reasonable fashion */

  /* FIXME: Keep track of the largest free chunk, and use this data */
  void * retval;
#ifdef DEBUG_VERY
  printf("chunk_alloc: Attempting to get %d (0x%x) bytes.\n", size, size);
#endif


  if(binrest)
    {
      retval = bin_get_best_fit(&binrest, size);
      if(retval)
	return conditional_split_chunk (retval, size);
      return 0;
    }
  
  return 0;

}

/* 
 * Grow the heap to allocate the requested chunk
 *
 * This function assumes that the heap needs to be grown
 *
 * Optimization note:
 *  If the top chunk is free, we just add to it to prevent fragmentation
 *  Always try and keep the top chunk free so this case doesn't introduce fragging
 */
void *grow_heap_alloc(int size)
{
  chunkptr top;
  int* top_size;
  int size_toalloc;
  int pages_toalloc;
  int * footer;

  /* check and see if top is free */
  
#ifdef DEBUG_BINCHECK
  if(binwalk_check())
    {
      printf("Bin boogered before attempt to grow_heap_alloc!\n");
      exit(0);
    }
#endif


  top_size = (int *) ((char *) dseg_hi -3) ;

#ifdef DEBUG_VERY
  printf("dseg_hi is 0x%x, looking for top's footer at 0x%x.  It indicates size of 0x%x.\n",
	(int) dseg_hi,
	(int) top_size,
	*top_size);
#endif

  if(!(*top_size & 0x1)) 
    {


      /* top is free, find where it begins */
      

      top = (chunkptr)  ( (char *) top_size - (char *) *top_size - 8);

      size_toalloc = (size - top->size + 7) & 0xFFFFFFF8;

      if(size_toalloc <= 0)
	{
	  printf("**** Top is free, and appears to be ");
	  printchunk(top);
	  printf("**** grow_alloc error: trying to allocate only %d bytes; doing binwalk: !\n",
		 size_toalloc);
	  binwalk();
	  if(!binwalk_check())
	    {
	      printf("** Error: Binwalk_check says there aren't any problems.  What gives?!?\n");
	    }

	  exit(0);
	}

           
#ifdef DEBUG_VERY 
      printf("Old heap:  dseg_hi=0x%x  dseg_lo=0x%x\n", (int) dseg_hi, (int) dseg_lo);	      
#endif

      mem_sbrk(size_toalloc);

      
#ifdef DEBUG_VERY
      printf("Heap grown:  dseg_hi=0x%x  dseg_lo=0x%x\n", (int) dseg_hi, (int) dseg_lo);
#endif

      /* FIXME: page size is a power of two.  What can we do to get rid of the multiply? */
      /* need not compensate for overhead bytes, since they weren't included in the old top->size */
      
      top->size = (top->size + (size_toalloc)) | 0x1; 

      footer = foot(top);
      *footer = top->size;
      

      /* remove top from the bins */

      /* delink */
      top->prev->next = top->next;
      top->next->prev = top->prev;

      
      /* FIXME: better way to do this?  Check and make sure this is right */

      /* if top is at the front of a bin, do necessary stuff */

      if(binrest == top)
	{
	  if(top->next == top)
	    {
	      binrest = NULL;
	    }
	  else
	    {
	      binrest = top->next;
	    }
	}


#ifdef DEBUG_BINCHECK

      if(binwalk_check())
	{
	  printf("** ERROR: Bin boogered after a grow_alloc (case 1)\n");
	  exit(0);
	}
#endif

      return top;
    }
  
  /* top is allocated, expand the heap and create a new one */

  {
   

    size_toalloc = (size + 12 + 7) & 0xFFFFFFF8;

//    pages_toalloc = (size + 12 + page_size -1 ) / page_size;
  

    /* the start of top is where the old top's footer is */
    top = (chunkptr) top_size;
    mem_sbrk(size_toalloc);
 
#ifdef DEBUG_VERY
   printf("New dseg_hi is at 0x%x\n", (int) dseg_hi);
#endif    

    /* make sure we mark the new top as allocated */

    top->size = (size_toalloc - 8) | 0x1;
    
    /* set the footer properly */
    footer = foot(top);
    *footer = top->size;
    
    
#ifdef DEBUG_VERY
    printf("grow_alloc: Grew new top: ");
    printchunk(top);

    /* the footer should be at the new end of the data segment.  check to see that it is */
    if( (char *) footer != (char *) (dseg_hi-3))
      printf("ERROR: %s:%d Footer not at end of heap!\n", __FILE__, __LINE__);
#endif
    
  
    /* if we allocated more than was needed, clean up by splitting
     */
    if(top)
      {

#ifdef DEBUG_BINCHECK
	if(binwalk_check())
	  {
	    printf("** ERROR: Bin boogered after a grow_alloc (case 1)\n");
	    exit(0);
	  }
#endif

	return conditional_split_chunk( top,size);

      }
    else
      return 0;
  }
}


int mm_init (void)
{
  chunkptr top;
  int * footer;

  /* initialize the bins */
  binrest = NULL;


  /* grab a page */
  page_size = mem_pagesize();
  mem_sbrk(page_size +4 );

  /* create the "top" chunk to fill all the memory that just got allocated */

#ifdef DEBUG
  printf("Creating initial top at 0x%x.\n", (int) dseg_lo);
#endif

  top = (chunkptr) dseg_lo;
  top->size =  page_size - 8; 
  top->next=top;
  top->prev=top;
  footer = foot(top);
  *footer = top->size;

#ifdef DEBUG_VERY
  printf("footer of top block at 0x%x set to size of 0x%x.\n",
	 (int) footer,
	 top->size);

#endif

  /* Add the top chunk to a bin */

#ifdef DEBUG_VERY
  printf("mm_init: (after allocation) dseg_hi: 0x%x, dseg_lo: 0x%x\n",
	 (int) dseg_hi,
	 (int) dseg_lo);


  printf("mm_init created top:");
  printchunk(top);

#endif

  chunk_free(top);
  
  return 0;

}



void *mm_malloc (size_t size)
{
  void * retval;

  if (size == 112) size = 128;


#ifdef DEBUG_BINCHEC
  if(binwalk_check())
    {
      printf("** ERROR: Boogered bin at some point.  Failing.\n");
      exit(0);
    }
#endif

#ifdef DEBUG_VERY
  printf("----- mm_malloc(%d)------\n", size);
#endif


  /* see if we can allocate out of existing bins */
  if( (retval = chunk_alloc(size)))
    {
#ifdef DEBUG_VERY
      printf("mm_mmalloc 0x%x (%d) bytes requested: Returning chunk:",
	     size,
	     size);
      printchunk((chunkptr) retval);
#endif

      retval = (void *) ( (char * ) retval + 8);
 
      return retval ;

    }

  /* Otherwise, grow the heap to allocate */

 if( (retval = grow_heap_alloc(size)))
    {
#ifdef DEBUG_VERY
      printf("mm_mmalloc 0x%x (%d) bytes requested: Returning chunk (heap grown):",
	     size,
	     size);
      printchunk((chunkptr) retval);
#endif
      

      retval = (void *) ( (char * ) retval + 8);
      

      return retval ;

    }

 return 0;
}


void mm_free (void *ptr)
{

#ifdef DEBUG_VERY
 printf("mm_free'ing 0x%x => 0x%x.\n", (int) ptr, (int) (((char *) ptr)- 8));
#endif

  if ( ((char*) ptr >= dseg_lo) && ( (char*) ptr <= dseg_hi))
    {
      chunkptr chunk = (chunkptr) (((char *) ptr)- 8);
     
      int * footer = foot(chunk);
  
      /* clear the allocated flag */
      *footer = chunk->size = chunk->size & 0xFFFFFFFE;
     
#ifdef COALESCING
      chunk = coalescing(chunk);
#endif
      
      chunk_free(chunk);
    }
  else
    {
#ifdef DEBUG
      printf("pointer being freed has invalid address!!!!!\n");
#endif
    }


}

