/* $Id$ */

/*
 *  Papadimitriou Spiros
 *  spapadim+@cs.cmu.edu
 *
 *  CS213 - Lab assignment 3
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <sys/mman.h>

#include <signal.h>
#include <asm/sigcontext.h>


#include "memlib.h"

char *dseg_lo = NULL, *dseg_hi;
long dseg_size;  /* Maximum size of data segment */

static int page_size;

/* Align pointer to closest page boundary downwards */
#define PAGE_ALIGN(p)    ((void *)(((unsigned long)(p) / page_size) * page_size))
/* Align pointer to closest page boundary upwards */
#define PAGE_ALIGN_UP(p) ((void *)((((unsigned long)(p) + page_size - 1) / page_size) * page_size))

#if defined(LINUX)
static void segv_handler (int sig, struct sigcontext sc)
#elif defined(ALPHA)
static void segv_handler(int sig, int code, struct sigcontext *scp, char *addr)
#endif
{
    char *addr;
    int  writing;

#if defined(LINUX)
    addr = (char *)sc.cr2;
    writing = sc.err & 0x02L;
#elif defined(ALPHA)
    addr = (char *)scp->sc_traparg_a0;
    writing = (*(unsigned *) scp->sc_pc & (1 << 28));
#endif

    fprintf(stderr, "Segmentation fault caught!\n");
    fprintf(stderr, "Address = %p\n", addr);
    fprintf(stderr, "Write = %s\n", (writing ? "yes" : "no"));
    fprintf(stderr, "Please use gdb to inspect the coredump!\n\n");
    abort();
}


int mem_init (void)
{
    struct sigaction segv_act;

    /* Get system page size */
    page_size = (int)getpagesize();

    /* Allocate heap */
    dseg_lo = (char *)malloc(DSEG_MAX + 2*page_size);
    if (!dseg_lo)
        return -1;
    dseg_lo = (char *)PAGE_ALIGN_UP(dseg_lo);

    /* Install SEGV handler */
    segv_act.sa_handler = (void (*)())segv_handler;
    sigemptyset(&segv_act.sa_mask);
    segv_act.sa_flags = 0;
    if (sigaction(SIGSEGV, &segv_act, NULL) < 0)
        return -1;

    return 0;
}

int mem_reinit (long size)
{
    dseg_size = (size > 0) ? size : DSEG_MAX;

    assert(dseg_lo != NULL);
    dseg_hi = dseg_lo - 1;

//  mprotect(PAGE_ALIGN(dseg_lo), DSEG_MAX, PROT_NONE);

    return 0;
}

void *mem_sbrk (ptrdiff_t increment)
{
    char *new_hi = dseg_hi + increment;
    char *old_hi = dseg_hi;
    long dseg_cursize = dseg_hi - dseg_lo + 1;

    assert(increment > 0);

    /* Resize data segment, if the memory is available */
    if (new_hi > dseg_lo + dseg_size)
        return NULL;
    dseg_hi = new_hi;
    dseg_cursize = dseg_hi - dseg_lo + 1;

    /* Properly protect unavailable memory
     * (note: order of mrotect()s matters) */
//  mprotect(PAGE_ALIGN(dseg_hi), DSEG_MAX - dseg_cursize, PROT_NONE);
//  mprotect(PAGE_ALIGN(dseg_lo), dseg_cursize, PROT_READ | PROT_WRITE);

    return (void *)(old_hi + 1);
}

int mem_pagesize (void)
{
    return page_size;
}

long mem_usage (void)
{
    return dseg_hi - dseg_lo;
}
 
